#include "bpe_tokenizer.h"

#include <cstdlib>
#include <fstream>
#include <nlohmann/json.hpp>
#include <stdexcept>

#include "log.h"

namespace fg {

// 定义 overloaded 类
template <class... Ts>
struct overloaded : Ts... {
  using Ts::operator()...;
};
// 类模板参数推导指引（C++17）
template <class... Ts>
overloaded(Ts...) -> overloaded<Ts...>;

std::string LoadBytesFromFile(const std::string& path) {
  std::ifstream fs(path, std::ios::in | std::ios::binary);
  if (fs.fail()) {
    LOG(ERROR) << "Cannot open " << path;
    throw std::runtime_error("Cannot open " + path);
  }
  std::string data;
  fs.seekg(0, std::ios::end);
  size_t size = static_cast<size_t>(fs.tellg());
  fs.seekg(0, std::ios::beg);
  data.resize(size);
  fs.read(data.data(), size);
  return data;
}

int BPETokenizer::Initialize(const string& feature_config) {
  nlohmann::json cfg;
  try {
    cfg = nlohmann::json::parse(feature_config);
  } catch (nlohmann::json::parse_error& ex) {
    LOG(ERROR) << "parse error at byte " << ex.byte;
    LOG(ERROR) << "config: " << feature_config;
    throw std::runtime_error("parse BPETokenizer config failed");
  }

  feature_name_ = cfg.at("feature_name");

  output_type_ = cfg.value("output_type", "word_id");
  if (output_type_ != "word" && output_type_ != "word_id") {
    string info("output_type should be one of [word, word_id]");
    ERROR_EXIT(feature_name_, info);
  }

  vocab_file_ = cfg.at("vocab_file");
  char* p = getenv("GLOBAL_FG_CONFIG_DIR");
  if (p && *p != '\0') {
    vocab_file_ = string(p) + "/" + vocab_file_;
    LOG(INFO) << feature_name_ << " vocab file " << vocab_file_;
  }
  auto blob = LoadBytesFromFile(vocab_file_);
  tokenizer_ = Tokenizer::FromBlobJSON(blob);
  if (nullptr == tokenizer_) {
    ERROR_EXIT(feature_name_, "tokenizer is null");
  }
  LOG(INFO) << "load model success. vocab size: " << vocab_size();

  dval_str_ = cfg.value("default_value", "");
  if (!dval_str_.empty()) {
    default_values_ = tokenizer_->Encode(dval_str_);
  }
  return 0;
}

int BPETokenizer::ProcessWithStrOutputs(const vector<Field>& inputs,
                                        vector<string>& outputs) {
  vector<int32> ids;
  int status = ProcessWithInt32Outputs(inputs, ids);
  if (0 != status) {
    return status;
  }
  outputs.clear();
  outputs.reserve(ids.size());
  if (output_type_ == "word_id") {
    std::transform(ids.begin(), ids.end(), std::back_inserter(outputs),
                   [](int32_t& x) { return std::to_string(x); });
  } else {
    for (auto& id : ids) {
      outputs.emplace_back(tokenizer_->IdToToken(id));
    }
  }
  return 0;
}

int BPETokenizer::ProcessWithInt32Outputs(const vector<Field>& inputs,
                                          vector<int32>& outputs) {
  outputs.clear();
  for (size_t i = 0; i < inputs.size(); ++i) {
    absl::visit(
        overloaded{
            [this, &outputs](const optional<string>& sentence) {
              if (!sentence.has_value() || sentence.value().empty()) {
                if (!dval_str_.empty()) {
                  outputs.insert(outputs.end(), default_values_.begin(),
                                 default_values_.end());
                }
              } else {
                // tokenize words ，get ids of words
                std::vector<int> ids = tokenizer_->Encode(sentence.value());
                outputs.insert(outputs.end(), ids.begin(), ids.end());
              }
            },
            [this, &outputs](const optional<absl::string_view>& sentence) {
              if (!sentence.has_value() || sentence.value().empty()) {
                if (!dval_str_.empty()) {
                  outputs.insert(outputs.end(), default_values_.begin(),
                                 default_values_.end());
                }
              } else {
                // tokenize words ，get ids of words
                std::vector<int> ids =
                    tokenizer_->Encode(string(sentence.value()));
                outputs.insert(outputs.end(), ids.begin(), ids.end());
              }
            },
            [this](auto x) {
              ERROR_EXIT(feature_name_,
                         "unsupported input type: ", typeid(x).name());
            }},
        inputs.at(i));
  }
  return 0;
}

int BPETokenizer::ProcessWithInt64Outputs(const vector<Field>& inputs,
                                          vector<int64>& outputs) {
  if (output_type_ != "word_id") {
    ERROR_EXIT(
        feature_name_,
        "output_type must be set to `word_id` when output type is int64");
  }
  vector<int32> ids;
  int status = ProcessWithInt32Outputs(inputs, ids);
  if (0 != status) {
    return status;
  }
  outputs.clear();
  outputs.insert(outputs.end(), ids.begin(), ids.end());
  return 0;
}

int BPETokenizer::ProcessWithFloatOutputs(const vector<Field>& inputs,
                                          vector<float>& outputs) {
  if (output_type_ != "word_id") {
    ERROR_EXIT(
        feature_name_,
        "output_type must be set to `word_id` when output type is float");
  }
  vector<int32> ids;
  int status = ProcessWithInt32Outputs(inputs, ids);
  if (0 != status) {
    return status;
  }
  outputs.clear();
  outputs.insert(outputs.end(), ids.begin(), ids.end());
  return 0;
}

int BPETokenizer::ProcessWithDoubleOutputs(const vector<Field>& inputs,
                                           vector<double>& outputs) {
  if (output_type_ != "word_id") {
    ERROR_EXIT(
        feature_name_,
        "output_type must be set to `word_id` when output type is double");
  }
  vector<int32> ids;
  int status = ProcessWithInt32Outputs(inputs, ids);
  if (0 != status) {
    return status;
  }
  outputs.clear();
  outputs.insert(outputs.end(), ids.begin(), ids.end());
  return 0;
}
}  // end of namespace fg